# -*- coding: utf-8 -*-
"""
    flask-sentinel.utils
    ~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""


class Config(object):
    def __init__(self, app):
        self.prefix = 'SENTINEL'
        self.app = app

        app.config.setdefault(self._key('MONGO_DBNAME'), 'oauth')
        app.config.setdefault(self._key('ROUTE_PREFIX'), '/oauth')
        app.config.setdefault(self._key('TOKEN_URL'), '/token')
        app.config.setdefault(self._key('MANAGEMENT_URL'), '/management')
        app.config.setdefault(self._key('REDIS_URL'),
                              'redis://localhost:6379/0')

    def url_rule_for(self, _key):
        return '%s%s' % (self.value('ROUTE_PREFIX'), self.value(_key))

    def value(self, key):
        return self.app.config[self._key(key)]

    def _key(self, _key):
        return '%s_%s' % (self.prefix, _key)



# -*- coding: utf-8 -*-
"""
    flask-sentinel.validator
    ~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
from flask_oauthlib.provider import OAuth2RequestValidator

from data import Storage


class MyRequestValidator(OAuth2RequestValidator):
    """ Defines a custom OAuth2 Request Validator based on the Client, User
        and Token models.
    """
    def __init__(self):
        self._clientgetter = Storage.get_client
        self._usergetter = Storage.get_user
        self._tokengetter = Storage.get_token
        self._tokensetter = Storage.save_token



# -*- coding: utf-8 -*-
"""
    flask-sentinel
    ~~~~~~~~~~~~~~

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
from flask import Blueprint

import views
from core import oauth, mongo, redis
from utils import Config
from validator import MyRequestValidator


class ResourceOwnerPasswordCredentials(object):
    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        config = Config(app)
        redis.from_url(config.value('REDIS_URL'))
        self.app.config['DEBUG'] = True
        self.register_blueprint(app)

        if config.value('TOKEN_URL') is not False:
            app.add_url_rule(
                config.url_rule_for('TOKEN_URL'),
                view_func=views.access_token,
                methods=['POST']
            )

        if config.value('MANAGEMENT_URL') is not False:
            app.add_url_rule(
                config.url_rule_for('MANAGEMENT_URL'),
                view_func=views.management,
                methods=['POST', 'GET']
            )

        mongo.init_app(app, config_prefix='SENTINEL_MONGO')
        oauth.init_app(app)
        oauth._validator = MyRequestValidator()

    def register_blueprint(self, app):
            module = Blueprint('flask-sentinel', __name__,
                               template_folder='templates')
            app.register_blueprint(module)
            return module



# -*- coding: utf-8 -*-
"""
    flask-sentinel.views
    ~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
from flask import render_template, request

from core import oauth
from data import Storage


@oauth.token_handler
def access_token(*args, **kwargs):
    """ This endpoint is for exchanging/refreshing an access token.

    Returns a dictionary or None as the extra credentials for creating the
    token response.

    :param *args: Variable length argument list.
    :param **kwargs: Arbitrary keyword arguments.
    """
    return None


def management():
    """ This endpoint is for vieweing and adding users and clients. """
    if request.method == 'POST' and request.form['submit'] == 'Add User':
        Storage.save_user(request.form['username'], request.form['password'])
    if request.method == 'POST' and request.form['submit'] == 'Add Client':
        Storage.generate_client()
    return render_template('management.html', users=Storage.all_users(),
                           clients=Storage.all_clients())



# -*- coding: utf-8 -*-
"""
    flask-sentinel.tests.tests
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
from base import TestBase
from flask_sentinel.core import mongo, redis
from flask_sentinel.data import Storage
from flask_sentinel.models import Client, User


class TestTokenEndpoint(TestBase):
    def test_methods_not_allowed(self):
        r = self.test_client.get(self.token_endpoint)
        self.assert405(r.status_code)

        r = self.test_client.put(self.token_endpoint)
        self.assert405(r.status_code)

        r = self.test_client.delete(self.token_endpoint)
        self.assert405(r.status_code)

        r = self.test_client.patch(self.token_endpoint)
        self.assert405(r.status_code)

    def test_invalid_token_request(self):
        r = self.test_client.post(self.token_endpoint)
        self.assert400(r.status_code)

        query = self.url % ('notreally', self.username, self.pw)
        r = self.test_client.post(query)
        self.assert401(r.status_code)

        query = self.url % (self.clientid, 'notreally', self.pw)
        r = self.test_client.post(query)
        self.assert401(r.status_code)

        query = self.url % (self.clientid, self.username, 'notreally')
        r = self.test_client.post(query)
        self.assert401(r.status_code)

    def test_valid_token_request(self):
        son = self.get_token()
        self.assertTrue('access_token' in son)
        self.assertTrue('token_type' in son)
        self.assertTrue('refresh_token' in son)
        self.assertTrue('scope' in son)
        self.assertEqual('Bearer', son['token_type'])
        self.assertEqual('', son['scope'])

        # Test that token has been stored in redis
        token_key = son['access_token']
        token = redis.get(token_key)
        # Test that key value matches user id.
        self.assertEqual(token, str(self.user.id))
        # Test that TTL has been set to custom value.
        ttl = redis.ttl(token_key)
        self.assertEqual(ttl, 999)


class TestAuthEndpoint(TestBase):
    def test_no_auth(self):
        r = self.test_client.get(self.auth_endpoint)
        self.assert401(r.status_code)

    def test_invalid_auth(self):
        headers = [('Authorization', 'Bearer DontThinkSo')]
        r = self.test_client.get(self.auth_endpoint, headers=headers)
        self.assert401(r.status_code)

        headers = [('Authorization', 'Basic DontThinkSo')]
        r = self.test_client.get(self.auth_endpoint, headers=headers)
        self.assert401(r.status_code)

    def test_valid_auth(self):
        son = self.get_token()
        token = son['access_token']
        headers = [('Authorization', 'Bearer %s' % token)]
        r = self.test_client.get(self.auth_endpoint, headers=headers)
        self.assert200(r.status_code)


class TestManagementEndpoint(TestBase):
    def test_man_endpoint(self):
        r = self.test_client.get(self.man_endpoint)
        self.assert200(r.status_code)


class TestStorage(TestBase):
    def test_get_client(self):
        client = Storage.get_client('notreally')
        self.assertIsNone(client)

        client = Storage.get_client(self.clientid)
        self.assertIsInstance(client, Client)
        self.assertEqual(client.client_id, self.clientid)
        self.assertEqual(client.client_type, 'public')
        self.assertEqual(client.allowed_grant_types, ['password'])
        self.assertEqual(client.default_scopes, [])
        self.assertEqual(client.default_redirect_uri, '')

    def test_generate_client(self):
        client = Storage.generate_client()
        self.assertEqual(mongo.db.clients.count(), 2)

        compare = Storage.get_client(client.client_id)
        self.assertEqual(client.client_id, compare.client_id)

    def test_all_clients(self):
        client = Storage.generate_client()
        clients = Storage.all_clients()

        self.assertIsInstance(clients, list)
        self.assertEqual(len(clients), 2)
        self.assertEqual(clients[0].client_id, self.clientid)
        self.assertEqual(clients[1].client_id, client.client_id)

    def test_get_user(self):
        user = Storage.get_user('notreally', 'notreally')
        self.assertIsNone(user)

        user = Storage.get_user(self.username, self.pw)
        self.assertIsInstance(user, User)
        self.assertEqual(user.username, self.username)
        self.assertEqual(user.hashpw, self.user.hashpw)

    def test_save_user(self):
        user = Storage.save_user('test', 'testpw')
        self.assertEqual(mongo.db.users.count(), 2)

        compare = Storage.get_user('test', 'testpw')
        self.assertEqual(compare.username, user.username)
        self.assertEqual(compare.hashpw, user.hashpw)

    def test_all_users(self):
        user = Storage.save_user('test', 'testpw')
        users = Storage.all_users()

        self.assertIsInstance(users, list)
        self.assertEqual(len(users), 2)
        self.assertEqual(users[0].username, self.user.username)
        self.assertEqual(users[0].hashpw, self.user.hashpw)
        self.assertEqual(users[1].username, user.username)
        self.assertEqual(users[1].hashpw, user.hashpw)



# -*- coding: utf-8 -*-
"""
    flask_sentinel.tests
    ~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
import simplejson as json
import unittest

from flask import Flask
from flask_sentinel import ResourceOwnerPasswordCredentials, oauth
from flask_sentinel.core import mongo
from flask_sentinel.data import Storage


class TestBase(unittest.TestCase):

    def setUp(self):
        self.app = Flask(__name__)
        self.dbkey = 'SENTINEL_MONGO_DBNAME'

        self.app.add_url_rule('/endpoint', view_func=restricted_access)
        self.app.config.update(self.settings())

        ResourceOwnerPasswordCredentials(self.app)

        self.context = self.app.test_request_context('/')
        self.context.push()
        self.test_client = self.app.test_client()

        mongo.cx.drop_database(self.app.config[self.dbkey])

        self.pw = 'pw'
        self.clientapp = Storage.generate_client()
        self.clientid = self.clientapp.client_id
        self.user = Storage.save_user("user", self.pw)
        self.username = self.user.username

        self.auth_endpoint = '/endpoint'
        self.token_endpoint = '/testauth/testtoken'
        self.man_endpoint = '/testauth/testman'
        self.url = '%s?%s' % (
            self.token_endpoint,
            'client_id=%s&grant_type=password&username=%s&password=%s'
        )

    def tearDown(self):
        mongo.cx.drop_database(self.app.config[self.dbkey])
        self.context.pop()

    def settings(self):
        return {
            self.dbkey: 'test_auth',
            'REDIS_URL': 'redis://localhost:6379/0',
            'OAUTH2_PROVIDER_TOKEN_EXPIRES_IN': 999,
            'SENTINEL_TOKEN_URL': '/testtoken',
            'SENTINEL_MANAGEMENT_URL': '/testman',
            'SENTINEL_ROUTE_PREFIX': '/testauth'
        }

    def get_token(self):
        query = self.url % (self.clientid, self.username, self.pw)
        r = self.test_client.post(query)
        self.assert200(r.status_code)
        return json.loads(r.get_data())

    def assert405(self, status_code):
        self.assertEqual(405, status_code)

    def assert400(self, status_code):
        self.assertEqual(400, status_code)

    def assert401(self, status_code):
        self.assertEqual(401, status_code)

    def assert200(self, status_code):
        self.assertEqual(200, status_code)


@oauth.require_oauth()
def restricted_access():
    """ This is an example endpoint we are trying to protect. """
    return "Congratulations, you made it through and accessed the protected " \
        "resource!"



# -*- coding: utf-8 -*-
"""
    flask-sentinel.data
    ~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
import inspect
from collections import namedtuple
from datetime import datetime, timedelta

import bcrypt
from werkzeug.security import gen_salt

from core import mongo, redis
from models import Client, User, Token


# TODO use SONManipulator instead of custom de/serializers perhaps?

# map collection _id (primary key) to class property.
idFieldsMap = namedtuple('idFields', 'cls, collection')
id = idFieldsMap(cls='id', collection='_id')


def _from_json(json, cls, as_list=False):
    """ Serializes a JSON stream to a list of objects, or a single objects
        if only a document is contained in the string.

    :param json: JSON dictionary or list.
    :parm cls: target class. Only properties matching json
               keys will be set to the corresponding json values.
    """
    if json is None:
        return None

    if not isinstance(json, list):
        json = [json]

    objs = []
    for json_item in json:
        obj = cls()

        properties, json_keys = set(_properties(obj)), set(json_item.keys())
        for property in set.intersection(properties, json_keys):
            try:
                setattr(obj, property, json_item[property])
            except AttributeError:
                # Probably a readonly property.
                pass

        if id.collection in json_item and json_item[id.collection] is not None:
            setattr(obj, id.cls, json_item[id.collection])

        objs.append(obj)

    return objs if as_list else (objs.pop() if len(objs) else None)


def _to_json(obj):
    """ Deserializes an object to a JSON stream.

    :param obj: object to be serialized to JSON.
    """

    json = {}
    for prop in _properties(obj):
        json[prop] = getattr(obj, prop)

    objid = getattr(obj, id.cls)
    if objid is not None:
        json[id.collection] = objid

    return json


def _properties(obj, include_id=False):
    """ Returns a list of object properties.

    :param obj: object to be inspected.
    :param include_id: True if db.cls (usually 'id') is to be included.
    """
    a = [
        name for (name, value) in inspect.getmembers(
            obj.__class__, lambda p: isinstance(p, property)
        ) if name != id.cls or include_id
    ]
    return a


class Storage(object):

    @staticmethod
    def get_client(client_id):
        """ Loads a client from mongodb and returns it as a Client or None.
        """
        json = mongo.db.clients.find_one({'client_id': client_id})
        return _from_json(json, Client)

    @staticmethod
    def get_user(username, password, *args, **kwargs):
        """ Loads a user from mongodb and returns it as a User or None.
        """
        user = mongo.db.users.find_one({'username': username})
        if user and password:
            encoded_pw = password.encode('utf-8')
            user_hash = user['hashpw'].encode('utf-8')
            user = mongo.db.users.find_one({
                'username': username,
                'hashpw': bcrypt.hashpw(encoded_pw, user_hash)
            })
        return _from_json(user, User)

    @staticmethod
    def get_token(access_token=None, refresh_token=None):
        """ Loads a token from mongob and returns it as a Token or None.
        """
        if not (access_token or refresh_token):
            return None

        if access_token:
            field, value = 'access_token', access_token
        elif refresh_token:
            field, value = 'refresh_token', refresh_token

        json = mongo.db.tokens.find_one({field: value})
        token = _from_json(json, Token)
        if token is None:
            return None

        json = mongo.db.users.find_one({id.collection: token.user_id})
        token.user = _from_json(json, User)

        return token

    @staticmethod
    def save_token(token, request, *args, **kwargs):
        client_id = request.client.client_id
        user_id = request.user.id

        # Make sure there is only one grant token for every (client, user)
        mongo.db.tokens.remove({'client_id': client_id, 'user_id': user_id})

        expires_in = token.pop('expires_in')
        expires = datetime.utcnow() + timedelta(seconds=expires_in)

        token = Token(
            client_id=request.client.client_id,
            user_id=user_id,
            token_type=token['token_type'],
            access_token=token['access_token'],
            refresh_token=token['refresh_token'],
            expires=expires,
        )

        # Add the access token to the Redis cache and set it to
        # expire at the appropriate time.
        redis.setex(token.access_token, expires_in, user_id)

        spec = {'user_id': user_id, 'client_id': client_id}

        # Replace token if it exists already, insert otherwise.
        mongo.db.tokens.update(spec, _to_json(token), upsert=True)

    @staticmethod
    def generate_client():
        client = Client()
        client.client_id = gen_salt(40)
        client.client_type = "public"
        mongo.db.clients.insert(_to_json(client))
        return client

    @staticmethod
    def save_user(username, password):
        salt = bcrypt.gensalt()
        hash = bcrypt.hashpw(password.encode('utf-8'), salt)
        user = User(username=username, hashpw=hash)
        user.id = mongo.db.users.insert(_to_json(user))
        return user

    @staticmethod
    def all_users():
        json = list(mongo.db.users.find())
        return _from_json(json, User, as_list=True)

    @staticmethod
    def all_clients():
        json = list(mongo.db.clients.find())
        return _from_json(json, Client, as_list=True)



# -*- coding: utf-8 -*-
"""
    flask-sentinel.models
    ~~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""


class BaseModel(object):
    def __init__(self, id=None):
        self._id = id

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value


class User(BaseModel):
    """ User which will be querying resources from the API.
    """
    def __init__(self, id=None, username=None, hashpw=None):
        super(User, self).__init__(id)
        self._username = username
        self._hashpw = hashpw

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        self._username = value

    @property
    def hashpw(self):
        return self._hashpw

    @hashpw.setter
    def hashpw(self, value):
        self._hashpw = value


class Client(BaseModel):
    """ Client application through which user is authenticating.

    RFC 6749 Section 2 (http://tools.ietf.org/html/rfc6749#section-2)
    describes clients:

     +----------+
     | Resource |
     |  Owner   |
     |          |
     +----------+
          v
          |    Resource Owner
         (A) Password Credentials
          |
          v
     +---------+                                  +---------------+
     |         |>--(B)---- Resource Owner ------->|               |
     |         |         Password Credentials     | Authorization |
     | Client  |                                  |     Server    |
     |         |<--(C)---- Access Token ---------<|               |
     |         |    (w/ Optional Refresh Token)   |               |
     +---------+                                  +---------------+

    Redirection URIs are mandatory for clients. We skip this requirement
    as this example only allows the resource owner password credentials
    grant (described in Section 4.3). In this flow, the Authorization
    Server will not redirect the user as described in subsection 3.1.2
    (Redirection Endpoint).
    """
    def __init__(self, id=None, client_id=None, client_type=None):
        super(Client, self).__init__(id)
        self._client_id = client_id
        self._client_type = client_type

    @property
    def client_id(self):
        return self._client_id

    @client_id.setter
    def client_id(self, value):
        self._client_id = value

    @property
    def client_type(self):
        return self._client_type

    @client_type.setter
    def client_type(self, value):
        self._client_type = value

    @property
    def allowed_grant_types(self):
        """ Returns allowed grant types.

        Presently, only the password grant type is allowed.
        """
        return ['password']

    @property
    def default_scopes(self):
        """ Returns default scopes associated with the Client. """
        return []

    @property
    def default_redirect_uri(self):
        """ Return a blank default redirect URI since we are not implementing
            redirects.
        """
        return ''


class Token(BaseModel):
    """ Access or refresh token

        Because of our current grant flow, we are able to associate tokens
        with the users who are requesting them. This can be used to track usage
        and potential abuse. Only bearer tokens currently supported.
    """

    def __init__(self, id=None, client_id=None, user_id=None, user=None,
                 token_type=None, access_token=None, refresh_token=None,
                 expires=None, scopes=['']):
        super(Token, self).__init__(id)
        self._client_id = client_id
        self._user_id = user_id
        self._user = None
        self._token_type = token_type
        self._access_token = access_token
        self._refresh_token = refresh_token
        self._expires = expires
        self._scopes = scopes

    @property
    def client_id(self):
        return self._client_id

    @client_id.setter
    def client_id(self, value):
        self._client_id = value

    @property
    def user_id(self):
        return self._user_id

    @user_id.setter
    def user_id(self, value):
        self._user_id = value

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, value):
        self._user = value

    @property
    def token_type(self):
        return self._token_type

    @token_type.setter
    def token_type(self, value):
        self._token_type = value

    @property
    def access_token(self):
        return self._access_token

    @access_token.setter
    def access_token(self, value):
        self._access_token = value

    @property
    def refresh_token(self):
        return self._refresh_token

    @refresh_token.setter
    def refresh_token(self, value):
        self._refresh_token = value

    @property
    def expires(self):
        return self._expires.replace(tzinfo=None)

    @expires.setter
    def expires(self, value):
        self._expires = value

    @property
    def scopes(self):
        return self._scopes

    @scopes.setter
    def scopes(self, value):
        self._scopes = value



#!/usr/bin/env python

from setuptools import setup, find_packages
DESCRIPTION = ('OAuth2 Provider for Flask applications.')
with open('README.rst') as f:
    LONG_DESCRIPTION = f.read()

install_requires = [
    'Flask-OAuthlib',
    'Flask-PyMongo',
    'bcrypt',
    'pyOpenSSL',
    'redis',
]

setup(
    name='Flask-Sentinel',
    version='0.0.2',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author='Nicola Iarocci',
    author_email='nicola@nicolaiarocci.com',
    url='https://github.com/nicolaiarocci/flask-sentinel',
    license='BSD',
    platforms=["any"],
    packages=find_packages(),
    test_suite="flask-sentinel.tests",
    install_requires=install_requires,
    tests_require=['redis'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



# -*- coding: utf-8 -*-
"""
    flask-sentinel.core
    ~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
from flask.ext.pymongo import PyMongo
from flask_oauthlib.provider import OAuth2Provider
from redis import StrictRedis

mongo = PyMongo()
oauth = OAuth2Provider()
redis = StrictRedis()
